import abc
from lark import Transformer, v_args
from typing import Dict
from optimade.server.mappers import BaseResourceMapper

__all__ = ("BaseTransformer",)


class BaseTransformer(abc.ABC, Transformer):
    """Generic filter transformer that handles various
    parts of the grammar in a backend non-specific way.

    """

    # map from standard comparison operators to the backend-specific version
    operator_map: Dict[str, str] = {
        "<": None,
        "<=": None,
        ">": None,
        ">=": None,
        "!=": None,
        "=": None,
    }

    # map from back-end specific operators to their inverse
    # e.g. {"$lt": "$gt"} for MongoDB.
    reversed_operator_map: Dict[str, str] = {}

    def __init__(self, mapper: BaseResourceMapper = None):
        """Initialise the transformer object, optionally loading in a
        resource mapper for use when post-processing.

        """
        self.mapper = mapper

    def postprocess(self, query):
        """Post-process the query according to the rules defined for
        the backend.

        """
        return query

    def transform(self, tree):
        """ Transform the query using the Lark transformer then post-process. """
        return self.postprocess(super().transform(tree))

    def __default__(self, data, children, meta):
        raise NotImplementedError(
            f"Calling __default__, i.e., unknown grammar concept. data: {data}, children: {children}, meta: {meta}"
        )

    def filter(self, arg):
        """ filter: expression* """
        return arg[0] if arg else None

    @v_args(inline=True)
    def constant(self, value):
        """ constant: string | number """
        # Note: Return as is.
        return value

    @v_args(inline=True)
    def value(self, value):
        """ value: string | number | property """
        # Note: Return as is.
        return value

    @v_args(inline=True)
    def non_string_value(self, value):
        """ non_string_value: number | property """
        # Note: Return as is.
        return value

    @v_args(inline=True)
    def not_implemented_string(self, value):
        """not_implemented_string: value

        Raises:
            NotImplementedError: For further information, see Materials-Consortia/OPTIMADE issue 157:
                https://github.com/Materials-Consortia/OPTIMADE/issues/157

        """
        raise NotImplementedError("Comparing strings is not yet implemented.")

    def property(self, arg):
        """ property: IDENTIFIER ( "." IDENTIFIER )* """
        return ".".join(arg)

    @v_args(inline=True)
    def string(self, string):
        """ string: ESCAPED_STRING """
        return string.strip('"')

    @v_args(inline=True)
    def signed_int(self, number):
        """ signed_int : SIGNED_INT """
        return int(number)

    @v_args(inline=True)
    def number(self, number):
        """ number: SIGNED_INT | SIGNED_FLOAT """
        if number.type == "SIGNED_INT":
            type_ = int
        elif number.type == "SIGNED_FLOAT":
            type_ = float
        return type_(number)

    @v_args(inline=True)
    def comparison(self, value):
        """ comparison: constant_first_comparison | property_first_comparison """
        # Note: Return as is.
        return value

    @abc.abstractmethod
    def value_list(self, arg):
        """ value_list: [ OPERATOR ] value ( "," [ OPERATOR ] value )* """

    @abc.abstractmethod
    def value_zip(self, arg):
        """ value_zip: [ OPERATOR ] value ":" [ OPERATOR ] value (":" [ OPERATOR ] value)* """

    @abc.abstractmethod
    def value_zip_list(self, arg):
        """ value_zip_list: value_zip ( "," value_zip )* """

    @abc.abstractmethod
    def expression(self, arg):
        """ expression: expression_clause ( OR expression_clause ) """

    @abc.abstractmethod
    def expression_clause(self, arg):
        """ expression_clause: expression_phrase ( AND expression_phrase )* """

    @abc.abstractmethod
    def expression_phrase(self, arg):
        """ expression_phrase: [ NOT ] ( comparison | "(" expression ")" ) """

    @abc.abstractmethod
    def property_first_comparison(self, arg):
        """property_first_comparison: property ( value_op_rhs | known_op_rhs | fuzzy_string_op_rhs | set_op_rhs |
        set_zip_op_rhs | length_op_rhs )

        """

    @abc.abstractmethod
    def constant_first_comparison(self, arg):
        """ constant_first_comparison: constant OPERATOR ( non_string_value | not_implemented_string ) """

    @v_args(inline=True)
    @abc.abstractmethod
    def value_op_rhs(self, operator, value):
        """ value_op_rhs: OPERATOR value """

    @abc.abstractmethod
    def known_op_rhs(self, arg):
        """ known_op_rhs: IS ( KNOWN | UNKNOWN ) """

    @abc.abstractmethod
    def fuzzy_string_op_rhs(self, arg):
        """ fuzzy_string_op_rhs: CONTAINS value | STARTS [ WITH ] value | ENDS [ WITH ] value """

    @abc.abstractmethod
    def set_op_rhs(self, arg):
        """ set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ANY value_list | ONLY value_list ) """

    @abc.abstractmethod
    def length_op_rhs(self, arg):
        """ length_op_rhs: LENGTH [ OPERATOR ] value """

    @abc.abstractmethod
    def set_zip_op_rhs(self, arg):
        """set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list |
        ANY value_zip_list )

        """

    @abc.abstractmethod
    def property_zip_addon(self, arg):
        """ property_zip_addon: ":" property (":" property)* """



""" This module implements filter transformer classes for different backends. These
classes typically parse the filter with Lark and produce an appropriate query for the
given backend.

"""

from optimade.filtertransformers.base_transformer import BaseTransformer

__all__ = ("BaseTransformer",)



import traceback
from typing import List, Tuple, Callable

from lark.exceptions import VisitError

from pydantic import ValidationError
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError, StarletteHTTPException
from fastapi import Request
from fastapi.responses import JSONResponse

from optimade.models import OptimadeError, ErrorResponse, ErrorSource

from optimade.server.config import CONFIG
from optimade.server.logger import LOGGER
from optimade.server.routers.utils import meta_values


def general_exception(
    request: Request,
    exc: Exception,
    status_code: int = 500,  # A status_code in `exc` will take precedence
    errors: List[OptimadeError] = None,
) -> JSONResponse:
    debug_info = {}
    if CONFIG.debug:
        tb = "".join(
            traceback.format_exception(etype=type(exc), value=exc, tb=exc.__traceback__)
        )
        LOGGER.error("Traceback:\n%s", tb)
        debug_info[f"_{CONFIG.provider.prefix}_traceback"] = tb

    try:
        http_response_code = int(exc.status_code)
    except AttributeError:
        http_response_code = int(status_code)

    try:
        title = str(exc.title)
    except AttributeError:
        title = str(exc.__class__.__name__)

    try:
        detail = str(exc.detail)
    except AttributeError:
        detail = str(exc)

    if errors is None:
        errors = [OptimadeError(detail=detail, status=http_response_code, title=title)]

    response = ErrorResponse(
        meta=meta_values(
            url=request.url,
            data_returned=0,
            data_available=0,
            more_data_available=False,
            **debug_info,
        ),
        errors=errors,
    )

    return JSONResponse(
        status_code=http_response_code,
        content=jsonable_encoder(response, exclude_unset=True),
    )


def http_exception_handler(request: Request, exc: StarletteHTTPException):
    return general_exception(request, exc)


def request_validation_exception_handler(request: Request, exc: RequestValidationError):
    return general_exception(request, exc)


def validation_exception_handler(request: Request, exc: ValidationError):
    status = 500
    title = "ValidationError"
    errors = set()
    for error in exc.errors():
        pointer = "/" + "/".join([str(_) for _ in error["loc"]])
        source = ErrorSource(pointer=pointer)
        code = error["type"]
        detail = error["msg"]
        errors.add(
            OptimadeError(
                detail=detail, status=status, title=title, source=source, code=code
            )
        )
    return general_exception(request, exc, status_code=status, errors=list(errors))


def grammar_not_implemented_handler(request: Request, exc: VisitError):
    rule = getattr(exc.obj, "data", getattr(exc.obj, "type", str(exc)))

    status = 501
    title = "NotImplementedError"
    detail = (
        f"Error trying to process rule '{rule}'"
        if not str(exc.orig_exc)
        else str(exc.orig_exc)
    )
    error = OptimadeError(detail=detail, status=status, title=title)
    return general_exception(request, exc, status_code=status, errors=[error])


def not_implemented_handler(request: Request, exc: NotImplementedError):
    status = 501
    title = "NotImplementedError"
    detail = str(exc)
    error = OptimadeError(detail=detail, status=status, title=title)
    return general_exception(request, exc, status_code=status, errors=[error])


def general_exception_handler(request: Request, exc: Exception):
    return general_exception(request, exc)


OPTIMADE_EXCEPTIONS: Tuple[Exception, Callable[[Request, Exception], JSONResponse]] = (
    (StarletteHTTPException, http_exception_handler),
    (RequestValidationError, request_validation_exception_handler),
    (ValidationError, validation_exception_handler),
    (VisitError, grammar_not_implemented_handler),
    (NotImplementedError, not_implemented_handler),
    (Exception, general_exception_handler),
)
"""A tuple of all pairs of exceptions and handler functions that allow for
appropriate responses to be returned in certain scenarios according to the
OPTIMADE specification.

To use these in FastAPI app code:

```python
from fastapi import FastAPI
app = FastAPI()
for exception, handler in OPTIMADE_EXCEPTIONS:
    app.add_exception_handler(exception, handler)
```

"""



import warnings

warnings.warn(
    "Django functionality is deprecated and will be removed in later versions (unless support is requested).",
    DeprecationWarning,
    stacklevel=2,
)

import operator
from optimade.filterparser import LarkParser
from lark import Tree
from lark.lexer import Token
from django.db.models import Q


__all__ = ("DjangoTransformer",)


class DjangoQueryError(Exception):
    pass


django_db_keys = {
    "natoms": "entry__natoms",
    "chemical_formula": "composition__formula__in",
    "stability": "stability",
    "band_gap": "calculation__band_gap",
    "element": "composition__element_list__contains",
    "nelements": "entry__composition__ntypes",
}


class DjangoTransformer:
    """Filter transformer for implementations using Django.

    !!! warning "Warning"
        This transformer is deprecated as it only supports
        the 0.9.7 grammar version, and works different to other
        filter transformers in this package.

    """

    def __init__(self):
        self.opers = {
            "=": self.eq,
            ">": self.gt,
            ">=": self.ge,
            "<": self.lt,
            "<=": self.le,
            "!=": self.ne,
            "OR": self.or_,
            "AND": self.and_,
            "NOT": self.not_,
        }
        self.parser = LarkParser(version=(0, 9, 7))

    def parse_raw_q(self, raw_query):
        return self.parser.parse(raw_query)

    def eq(self, a, b):
        return Q(**{a: b})

    def gt(self, a, b):
        return Q(**{a + "__gt": b})

    def ge(self, a, b):
        return Q(**{a + "__gte": b})

    def lt(self, a, b):
        return Q(**{a + "__lt": b})

    def le(self, a, b):
        return Q(**{a + "__lte": b})

    def ne(self, a, b):
        return ~Q(**{a: b})

    def not_(self, a):
        return ~a

    def and_(self, a, b):
        return operator.and_(a, b)

    def or_(self, a, b):
        return operator.or_(a, b)

    def evaluate(self, parse_Tree):
        if isinstance(parse_Tree, Tree):
            children = parse_Tree.children
            if len(children) == 1:
                return self.evaluate(children[0])
            elif len(children) == 2:
                op_fn = self.evaluate(children[0])
                return op_fn(self.evaluate(children[1]))
            elif len(children) == 3:
                if parse_Tree.data == "comparison":
                    db_prop = self.evaluate(children[0])
                    op_fn = self.evaluate(children[1])

                    if db_prop in django_db_keys.keys():
                        return op_fn(
                            django_db_keys[db_prop], self.evaluate(children[2])
                        )
                    else:
                        raise DjangoQueryError(
                            "Unknown property is queried : " + (db_prop)
                        )

                else:
                    op_fn = self.evaluate(children[1])
                    return op_fn(self.evaluate(children[0]), self.evaluate(children[2]))
            else:
                raise DjangoQueryError("Not compatible format. Tree has >3 children")

        elif isinstance(parse_Tree, Token):
            if parse_Tree.type == "VALUE":
                return parse_Tree.value
            elif parse_Tree.type in ["NOT", "CONJUNCTION", "OPERATOR"]:
                return self.opers[parse_Tree.value]
        else:
            raise DjangoQueryError("Not a Lark Tree or Token")



from fastapi import HTTPException

__all__ = (
    "BadRequest",
    "VersionNotSupported",
    "Forbidden",
)


class StrReprMixin(HTTPException):
    """This mixin can be useful when testing requires a string
    representation of an exception that contains the HTTPException
    detail string, rather than the standard Python exception message.
    """

    def __str__(self):
        return self.__repr__()


class BadRequest(StrReprMixin, HTTPException):
    """400 Bad Request"""

    def __init__(
        self,
        status_code: int = 400,
        detail: str = None,
        headers: dict = None,
        title: str = "Bad Request",
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)
        self.title = title


class VersionNotSupported(StrReprMixin, HTTPException):
    """553 Version Not Supported"""

    def __init__(
        self,
        status_code: int = 553,
        detail: str = None,
        headers: dict = None,
        title: str = "Version Not Supported",
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)
        self.title = title


class Forbidden(StrReprMixin, HTTPException):
    """403 Forbidden"""

    def __init__(
        self,
        status_code: int = 403,
        detail: str = None,
        headers: dict = None,
        title: str = "Forbidden",
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)
        self.title = title



import os

from typing import Tuple, List, Union
from fastapi import HTTPException

from optimade.filterparser import LarkParser
from optimade.filtertransformers.mongo import MongoTransformer
from optimade.models import EntryResource
from optimade.server.config import CONFIG
from optimade.server.entry_collections import EntryCollection
from optimade.server.logger import LOGGER
from optimade.server.mappers import BaseResourceMapper
from optimade.server.query_params import EntryListingQueryParams, SingleEntryQueryParams

try:
    CI_FORCE_MONGO = bool(int(os.environ.get("OPTIMADE_CI_FORCE_MONGO", 0)))
except (TypeError, ValueError):  # pragma: no cover
    CI_FORCE_MONGO = False


if CONFIG.use_real_mongo or CI_FORCE_MONGO:
    from pymongo.collection import Collection
    from pymongo import MongoClient

    client = MongoClient(CONFIG.mongo_uri)
    LOGGER.info("Using: Real MongoDB (pymongo)")
else:
    from mongomock.collection import Collection
    from mongomock import MongoClient

    client = MongoClient()
    LOGGER.info("Using: Mock MongoDB (mongomock)")


class MongoCollection(EntryCollection):
    """Class for querying MongoDB collections (implemented by either pymongo or mongomock)
    containing serialized [`EntryResource`][optimade.models.entries.EntryResource]s objects.

    """

    def __init__(
        self,
        collection: Collection,
        resource_cls: EntryResource,
        resource_mapper: BaseResourceMapper,
    ):
        """Initialize the MongoCollection for the given parameters.

        Parameters:
            collection (Union[pymongo.collection.Collection, mongomock.collection.Collection]):
                The backend-specific collection.
            resource_cls (EntryResource): The `EntryResource` model
                that is stored by the collection.
            resource_mapper (BaseResourceMapper): A resource mapper
                object that handles aliases and format changes between
                deserialization and response.

        """
        super().__init__(
            collection,
            resource_cls,
            resource_mapper,
            MongoTransformer(mapper=resource_mapper),
        )

        self.parser = LarkParser(version=(1, 0, 0), variant="default")

        # check aliases do not clash with mongo operators
        self._check_aliases(self.resource_mapper.all_aliases())
        self._check_aliases(self.resource_mapper.all_length_aliases())

    def __len__(self) -> int:
        return self.collection.estimated_document_count()

    def count(self, **kwargs) -> int:
        """Returns the number of entries matching the query specified
        by the keyword arguments.

        Parameters:
            kwargs (dict): Query parameters as keyword arguments. The keys
                'filter', 'skip', 'limit', 'hint' and 'maxTimeMS' will be passed
                to the `pymongo.collection.Collection.count_documents` method.

        """
        for k in list(kwargs.keys()):
            if k not in ("filter", "skip", "limit", "hint", "maxTimeMS"):
                del kwargs[k]
        if "filter" not in kwargs:  # "filter" is needed for count_documents()
            kwargs["filter"] = {}
        return self.collection.count_documents(**kwargs)

    def find(
        self, params: Union[EntryListingQueryParams, SingleEntryQueryParams]
    ) -> Tuple[List[EntryResource], int, bool, set]:
        """Perform the query on the underlying MongoCollection, handling projection
        and pagination of the output.

        Returns:
            Tuple[List[EntryResource], int, bool, set]: A list of entry resource objects, the number of returned entries,
            whether more are available with pagination, fields.

        """

        criteria = self.handle_query_params(params)

        all_fields = criteria.pop("fields")
        if getattr(params, "response_fields", False):
            fields = set(params.response_fields.split(","))
            fields |= self.resource_mapper.get_required_fields()
        else:
            fields = all_fields.copy()

        results = []
        for doc in self.collection.find(**criteria):
            if criteria.get("projection", {}).get("_id"):
                doc["_id"] = str(doc["_id"])
            results.append(self.resource_cls(**self.resource_mapper.map_back(doc)))

        nresults_now = len(results)
        if isinstance(params, EntryListingQueryParams):
            criteria_nolimit = criteria.copy()
            criteria_nolimit.pop("limit", None)
            data_returned = self.count(**criteria_nolimit)
            more_data_available = nresults_now < data_returned
        else:
            # SingleEntryQueryParams, e.g., /structures/{entry_id}
            data_returned = nresults_now
            more_data_available = False
            if nresults_now > 1:
                raise HTTPException(
                    status_code=404,
                    detail=f"Instead of a single entry, {nresults_now} entries were found",
                )
            results = results[0] if results else None

        return results, data_returned, more_data_available, all_fields - fields

    def _check_aliases(self, aliases):
        """ Check that aliases do not clash with mongo keywords. """
        if any(
            alias[0].startswith("$") or alias[1].startswith("$") for alias in aliases
        ):
            raise RuntimeError(f"Cannot define an alias starting with a '$': {aliases}")
